import dataclasses
import shlex
import shutil
import subprocess
from pathlib import Path
from typing import ClassVar, Dict, Iterable, List, Tuple, Union

from ._error import Error
from .compat import StrEnum


class ExecutableNotFound(Error):

    def __init__(
        self, name_or_path: Union[str, Path], message: str,
    ) -> None:
        self.name_or_path = name_or_path
        self.message = message

    def __str__(self) -> str:
        return f'{self.name_or_path} not found: {self.message}'


class ToolVersionCheckFailed(Error):
    pass


class ToolNotFound(Error):

    def __init__(self, tool: str, candidates: Iterable[str]) -> None:
        self.tool = tool
        self.candidates = tuple(candidates)

    def __str__(self) -> str:
        candidates = ', '.join(f'`{c}`' for c in self.candidates)
        return (
            f'{self.tool} tool not found, '
            f'candidates tried: {candidates}'
        )


class Tool(StrEnum):
    COMPILE = 'compile'
    SYNC = 'sync'


@dataclasses.dataclass(frozen=True)
class Provider:
    PIP_TOOLS: ClassVar['Provider']
    UV: ClassVar['Provider']

    tools: Dict[Tool, str]

    _registry: ClassVar[Dict[str, 'Provider']] = {}

    @classmethod
    def register(cls, attr_name: str, provider: 'Provider') -> None:
        setattr(cls, attr_name, provider)
        cls._registry[attr_name] = provider

    @classmethod
    def get_tool_candidates(cls, tool: Union[Tool, str]) -> Tuple[str, ...]:
        if not isinstance(tool, Tool):
            tool = Tool(tool)
        return tuple(prov.tools[tool] for prov in cls._registry.values())


Provider.register('PIP_TOOLS', Provider(
    tools={
        Tool.COMPILE: 'pip-compile',
        Tool.SYNC: 'pip-sync',
    },
))


Provider.register('UV', Provider(
    tools={
        Tool.COMPILE: 'uv pip compile',
        Tool.SYNC: 'uv pip sync',
    },
))


def find_executable(name_or_path: Union[Path, str]) -> Path:
    if isinstance(name_or_path, str):
        # execname, but not /foo/bar/execname, bar/execname, ./execname, etc.
        if Path(name_or_path).name == name_or_path:
            if _exec_path := shutil.which(name_or_path):
                return Path(_exec_path)
            raise ExecutableNotFound(name_or_path, 'not in PATH')
        name_or_path = Path(name_or_path)
    exec_path = name_or_path.resolve()
    if not exec_path.exists():
        raise ExecutableNotFound(exec_path, 'does not exist')
    if not exec_path.is_file():
        raise ExecutableNotFound(exec_path, 'not a file')
    return exec_path


def process_command_line(command_line: Union[Iterable[str], str]) -> List[str]:
    if isinstance(command_line, str):
        # XXX: won't properly work on Windows!
        command_line = shlex.split(command_line)
    else:
        command_line = list(command_line)
    command_line[0] = str(find_executable(command_line[0]))
    return command_line


def check_tool_version(
    command_line: Union[Iterable[str], str],
) -> Tuple[List[str], str]:
    try:
        command_line = process_command_line(command_line)
    except ExecutableNotFound as exc:
        raise ToolVersionCheckFailed from exc
    try:
        output = subprocess.check_output(
            [*command_line, '--version'], stderr=subprocess.STDOUT, text=True)
    except subprocess.CalledProcessError as exc:
        raise ToolVersionCheckFailed from exc
    return command_line, output.strip()


def find_tool(tool: Union[Tool, str]) -> Tuple[List[str], str]:
    candidates = Provider.get_tool_candidates(tool)
    for command_line in candidates:
        try:
            return check_tool_version(command_line)
        except ToolVersionCheckFailed:
            pass
    raise ToolNotFound(tool, candidates)



from ._error import Error
from .commands.compile import CompileError
from .commands.sync import SyncError
from .infile import (
    CircularReference, InputDirectoryError, ReferenceError, UnknownReference,
)
from .providers import ExecutableNotFound, ToolNotFound, ToolVersionCheckFailed


__all__ = [
    'Error',
    'InputDirectoryError',
    'ReferenceError',
    'CircularReference',
    'UnknownReference',
    'ExecutableNotFound',
    'ToolNotFound',
    'ToolVersionCheckFailed',
    'CompileError',
    'SyncError',
]



import logging
import subprocess
from pathlib import Path
from typing import Iterable, Optional, Union

from .._error import Error
from ..infile import ReferenceType, get_infiles, get_input_dir
from ..utils import try_relative_to


log = logging.getLogger(__name__)


class CompileError(Error):
    pass


def compile(
    command_line: Iterable[Union[Path, str]],
    input_dir: Optional[Union[Path, str]] = None,
) -> None:
    log.debug('using %s', command_line)
    input_dir = get_input_dir(input_dir)
    log.debug('input dir: %s', input_dir)
    cwd = Path.cwd()
    for infile in get_infiles(input_dir):
        log.info('compiling %s', infile)
        output_file = try_relative_to(input_dir / infile.output_name, cwd)
        with infile.temporarily_write_to(
            input_dir, references_as=ReferenceType.CONSTRAINTS,
        ) as input_file:
            cmd = [
                *command_line,
                try_relative_to(input_file, cwd),
                '-o', output_file,
            ]
            log.debug('calling %s', cmd)
            try:
                subprocess.check_call(cmd)
            except subprocess.CalledProcessError as exc:
                raise CompileError from exc



import subprocess
from pathlib import Path
from typing import Optional


class Error(Exception):

    def get_error_message_from_cause(self) -> Optional[str]:
        if not (cause := self.__cause__):
            return None
        if isinstance(cause, subprocess.CalledProcessError):
            cmd = cause.cmd
            if not isinstance(cmd, (str, Path)):
                cmd = ' '.join(map(str, cmd))
            msg = f'{cmd} returned non-zero exit status {cause.returncode}'
            if output := cause.output:
                if isinstance(output, bytes):
                    output = output.decode()
                msg = f'{msg}:\n{output}'
            return msg
        return str(cause)

    def __str__(self) -> str:
        if self.args:
            return super().__str__()
        return self.get_error_message_from_cause() or ''



import logging
from pathlib import Path
from typing import Optional, Union

from ..infile import get_infiles, get_input_dir


log = logging.getLogger(__name__)


def show(input_dir: Optional[Union[Path, str]] = None) -> None:
    input_dir = get_input_dir(input_dir)
    log.debug('input dir: %s', input_dir)
    for infile in get_infiles(input_dir):
        print('#', infile.generated_name)
        print(infile.render())



import operator
import re
from contextlib import contextmanager
from pathlib import Path
from typing import (
    Any, Dict, Iterable, Iterator, List, Literal, Optional, Tuple, Union, cast,
)

from ._error import Error
from .compat import StrEnum


class InputDirectoryError(Error):
    pass


class ReferenceError(Error):
    pass


class UnknownReference(ReferenceError):
    pass


class CircularReference(ReferenceError):

    def __init__(self, infiles: Iterable['InFile']) -> None:
        self.infiles = tuple(infiles)

    def __str__(self) -> str:
        return ', '.join(map(str, self.infiles))


ReferenceTypeOrLiteral = Union['ReferenceType', Literal['r', 'c']]


class ReferenceType(StrEnum):
    REQUIREMENTS = 'r'
    CONSTRAINTS = 'c'

    @classmethod
    def cast(cls, value: ReferenceTypeOrLiteral) -> 'ReferenceType':
        if isinstance(value, cls):
            return value
        return cls(value)


class Reference:
    type: ReferenceType
    infile: 'InFile'

    def __init__(
        self, type: ReferenceTypeOrLiteral, infile: 'InFile',
    ) -> None:
        type = ReferenceType.cast(type)
        self.type = type
        self.infile = infile
        self._tuple = (type, infile)

    def __str__(self) -> str:
        return f'-{self.type} {self.infile.output_name}'

    def __repr__(self) -> str:   # pragma: no cover
        return f'{self.type.name.capitalize()}({self.infile})'

    def __eq__(self, other: Any) -> bool:   # type: ignore[misc]
        if not isinstance(other, self.__class__):
            return NotImplemented
        return self._tuple == other._tuple

    def __hash__(self) -> int:
        return hash(self._tuple)

    def copy_as(self, type: ReferenceTypeOrLiteral) -> 'Reference':
        type = ReferenceType.cast(type)
        return self.__class__(type, self.infile)


class InFile:
    original_name: str
    generated_name: str
    output_name: str
    stem: str
    references: List[Reference]
    dependencies: List[str]

    def __init__(self, name_or_path: Union[Path, str]) -> None:
        path = Path(name_or_path)
        self.stem = stem = path.stem
        self.original_name = path.name
        self.generated_name = f'{stem}.ptl.in'
        self.output_name = f'{stem}.txt'
        self.references = []
        self.dependencies = []

    def __str__(self) -> str:
        return self.original_name

    def __repr__(self) -> str:   # pragma: no cover
        return f'{self.__class__.__name__}({self})'

    def __eq__(self, other: Any) -> bool:   # type: ignore[misc]
        if not isinstance(other, self.__class__):
            return NotImplemented
        return self.original_name == other.original_name

    def __hash__(self) -> int:
        return hash(self.original_name)

    def add_reference(self, reference: Reference) -> None:
        self.references.append(reference)

    def iterate_references(
        self, *, recursive: bool,
        as_: Optional[ReferenceTypeOrLiteral] = None,
    ) -> Iterator[Reference]:
        if as_:
            as_ = ReferenceType.cast(as_)
        for ref in self.references:
            if as_:
                yield ref.copy_as(as_)
            else:
                yield ref
            if recursive:
                recursive_as: Optional[ReferenceType]
                if as_:
                    recursive_as = as_
                elif ref.type == ReferenceType.CONSTRAINTS:
                    recursive_as = ReferenceType.CONSTRAINTS
                else:
                    recursive_as = None
                yield from ref.infile.iterate_references(
                    recursive=True, as_=recursive_as)

    def add_dependency(self, dependency: str) -> None:
        self.dependencies.append(dependency.strip())

    def render(
        self, *, references_as: Optional[ReferenceTypeOrLiteral] = None,
    ) -> str:
        return ''.join(self._line_iter(references_as=references_as))

    def write_to(
        self, directory: Union[Path, str],
        *, references_as: Optional[ReferenceTypeOrLiteral] = None,
    ) -> Path:
        path = Path(directory) / self.generated_name
        with open(path, 'wt') as fobj:
            fobj.writelines(self._line_iter(references_as=references_as))
        return path

    @contextmanager
    def temporarily_write_to(
        self, directory: Union[Path, str],
        *, references_as: Optional[ReferenceTypeOrLiteral] = None,
    ) -> Iterator[Path]:
        path: Optional[Path] = None
        try:
            yield (path := self.write_to(
                directory, references_as=references_as))
        finally:
            if path is not None:
                path.unlink()

    def _line_iter(
        self, *, references_as: Optional[ReferenceTypeOrLiteral] = None,
    ) -> Iterator[str]:
        references = self.iterate_references(recursive=True)
        if references_as:
            references = (ref.copy_as(references_as) for ref in references)
        for reference in references:
            yield f'{reference}\n'
        for dependency in self.dependencies:
            yield f'{dependency}\n'


_INLINE_COMMENT_REGEX = re.compile(r'\s+#')
_REFERENCE_REGEX = re.compile(
    r'^-(?P<type>[rc])\s*(?P<stem>[\w.-]+?)(?:\.in|\.txt)?$')


def read_infiles(input_dir: Union[Path, str]) -> Tuple[InFile, ...]:
    input_dir = Path(input_dir)
    input_paths = tuple(input_dir.glob('*.in'))
    stems_to_infiles: Dict[str, InFile] = {
        (infile := InFile(path)).stem: infile for path in input_paths}
    for infile in stems_to_infiles.values():
        with open(input_dir / infile.original_name) as fobj:
            for line in fobj:
                line = line.strip()
                if not line or line.startswith('#'):
                    continue
                if comment_match := _INLINE_COMMENT_REGEX.search(line):
                    line = line[:comment_match.start()]
                ref_match = _REFERENCE_REGEX.fullmatch(line)
                if not ref_match:
                    infile.add_dependency(line)
                    continue
                ref_type = cast(Literal['r', 'c'], ref_match.group('type'))
                ref_stem: str = ref_match.group('stem')
                try:
                    ref_infile = stems_to_infiles[ref_stem]
                except KeyError:
                    raise UnknownReference(f'{infile}: {ref_stem}')
                ref = Reference(ref_type, ref_infile)
                infile.add_reference(ref)
    return tuple(stems_to_infiles.values())


def sort_infiles(infiles: Iterable[InFile]) -> List[InFile]:
    sorted_infiles: List[InFile] = []
    _infiles_with_refs: Dict[InFile, List[Reference]] = {
        infile: infile.references for infile in infiles}
    while True:
        _processed = {
            infile for infile, refs in _infiles_with_refs.items() if not refs}
        if not _processed:
            break
        sorted_infiles.extend(
            sorted(_processed, key=operator.attrgetter('stem')))
        _infiles_with_refs = {
            infile: [ref for ref in refs if ref.infile not in _processed]
            for infile, refs in _infiles_with_refs.items()
            if infile not in _processed
        }
    if len(_infiles_with_refs) != 0:
        raise CircularReference(_infiles_with_refs)
    return sorted_infiles


def get_infiles(input_dir: Union[Path, str]) -> List[InFile]:
    infiles = read_infiles(input_dir)
    if not infiles:
        raise InputDirectoryError('no *.in files')
    return sort_infiles(infiles)


def get_input_dir(input_dir: Optional[Union[Path, str]] = None) -> Path:
    if input_dir:
        input_dir = Path(input_dir)
        if not input_dir.exists():
            raise InputDirectoryError(f'{input_dir} does not exist')
        if not input_dir.is_dir():
            raise InputDirectoryError(f'{input_dir} is not a directory')
    else:
        input_dir = Path('requirements')
        if not input_dir.is_dir():
            input_dir = Path('.')
            if not next(input_dir.glob('*.in'), None):
                raise InputDirectoryError('input directory not found')
    return input_dir.resolve()



import textwrap
from pathlib import Path
from typing import Optional

import pytest


def dedent(text: str, *, strip: bool = True) -> str:
    if strip:
        text = text.lstrip('\n')
    return textwrap.dedent(text)


class InFileTestSuite:
    input_dir: Path

    @pytest.fixture(autouse=True)
    def base_setup(self, input_dir: Path) -> None:
        self.input_dir = input_dir

    dedent = staticmethod(dedent)

    def create_file(self, name: str, content: Optional[str] = None) -> Path:
        infile = self.input_dir / name
        if content:
            infile.write_text(self.dedent(content))
        else:
            infile.touch()
        return infile



import logging
import subprocess
from pathlib import Path
from typing import Iterable, List, Optional, Union

from .._error import Error
from ..infile import get_infiles, get_input_dir
from ..utils import try_relative_to


log = logging.getLogger(__name__)


class SyncError(Error):
    pass


def sync(
    command_line: Iterable[Union[Path, str]],
    input_dir: Optional[Union[Path, str]] = None,
) -> None:
    log.debug('using %s', command_line)
    input_dir = get_input_dir(input_dir)
    log.debug('input dir: %s', input_dir)
    cwd = Path.cwd()
    compiled_files: List[Path] = []
    missing_files: List[str] = []
    for infile in get_infiles(input_dir):
        output_name = infile.output_name
        compiled_file = input_dir / output_name
        if compiled_file.exists():
            compiled_files.append(try_relative_to(compiled_file, cwd))
        else:
            missing_files.append(output_name)
    if missing_files:
        raise SyncError(
            f'not all files are compiled, missing: {", ".join(missing_files)}')
    log.debug('syncing %s', compiled_files)
    cmd = [
        *command_line,
        *compiled_files,
    ]
    log.debug('calling %s', cmd)
    try:
        subprocess.check_call(cmd)
    except subprocess.CalledProcessError as exc:
        raise SyncError from exc



import argparse
import logging
import sys
from typing import List, Optional, Sequence

from . import __version__, commands
from .exceptions import Error
from .providers import (
    Provider, Tool, check_tool_version, find_tool, process_command_line,
)


log = logging.getLogger(__name__)


class Args(argparse.Namespace):
    command: str

    use_uv: bool
    use_pip_tools: bool
    custom_tool: Optional[str]

    input_dir: Optional[str]
    verbose: int
    quiet: int


def add_command_options(
    command: str, command_parser: argparse.ArgumentParser,
    add_tool_selection: bool = True,
) -> None:
    if add_tool_selection:
        _tool_selection = command_parser.add_argument_group('tool selection')
        tool_selection = _tool_selection.add_mutually_exclusive_group()
        tool_selection.add_argument(
            '--pip-tools', action='store_true', dest='use_pip_tools',
            help=f'use `pip-{command}`',
        )
        tool_selection.add_argument(
            '--uv', action='store_true', dest='use_uv',
            help=f'use `uv pip {command}`',
        )
        tool_selection.add_argument(
            '--tool', metavar='TOOL', dest='custom_tool',
            help='use custom tool',
        )

    command_options = command_parser.add_argument_group(f'{command} options')
    command_options.add_argument(
        '-d', '--directory', metavar='DIR', dest='input_dir',
        help='input directory',
    )

    logging_options = command_options.add_mutually_exclusive_group()
    logging_options.add_argument(
        '-v', '--verbose', action='count', default=0,
        help='get more output',
    )
    logging_options.add_argument(
        '-q', '--quiet', action='count', default=0,
        help='get less output',
    )

    general_options = command_parser.add_argument_group('general options')
    general_options.add_argument(
        '-h', '--help', action='help',
        help='show this help message and exit',
    )


def build_parser() -> argparse.ArgumentParser:
    root_parser = argparse.ArgumentParser(prog='ptl', add_help=False)
    root_general_options = root_parser.add_argument_group('general options')
    root_general_options.add_argument(
        '-h', '--help', action='help',
        help='show this help message and exit',
    )
    root_general_options.add_argument(
        '-V', '--version', action='version', version=__version__,
        help='show version and exit',
    )

    subparsers = root_parser.add_subparsers(
        title='Commands', required=True, dest='command', metavar='COMMAND')

    compile_parser = subparsers.add_parser(
        'compile', add_help=False,
        help='compile requirements',
    )
    add_command_options('compile', compile_parser)

    sync_parser = subparsers.add_parser(
        'sync', add_help=False,
        help='sync requirements',
    )
    add_command_options('sync', sync_parser)

    show_parser = subparsers.add_parser(
        'show', add_help=False,
        help='show requirements',
    )
    add_command_options('show', show_parser, add_tool_selection=False)

    return root_parser


def main(argv: Optional[Sequence[str]] = None) -> None:
    try:
        do_main(argv)
    except Error as exc:
        log.error('%s: %s', exc.__class__.__name__, exc)
        sys.exit(1)


def do_main(argv: Optional[Sequence[str]] = None) -> None:
    parser = build_parser()
    args, tool_args = parser.parse_known_args(argv, namespace=Args())
    command = args.command

    is_tool: bool
    try:
        Tool(command)
    except ValueError:
        is_tool = False
    else:
        is_tool = True

    if not is_tool and tool_args:
        # only sync and compile accept extra args, other commands should raise
        # an error, the simplest way to do it is just call parse_args()
        parser.parse_args(argv)

    verbosity: int = 0
    verbosity_arg: Optional[str] = None
    if quiet := args.quiet:
        verbosity = -quiet
        verbosity_arg = '-{}'.format('q' * quiet)
    elif verbose := args.verbose:
        verbosity = verbose
        verbosity_arg = '-{}'.format('v' * verbose)
    configure_logging(verbosity)

    input_dir = args.input_dir

    if is_tool:
        tool_command_line = get_tool_command_line(args)
        if verbosity_arg:
            tool_command_line.append(verbosity_arg)
        tool_command_line.extend(tool_args)
        if command == Tool.COMPILE:
            commands.compile(
                command_line=tool_command_line,
                input_dir=input_dir,
            )
        elif command == Tool.SYNC:
            commands.sync(
                command_line=tool_command_line,
                input_dir=input_dir,
            )
        else:
            assert False, 'should not reach here'
    elif command == 'show':
        commands.show(input_dir)
    else:
        assert False, 'should not reach here'


def configure_logging(verbosity: int) -> None:
    if verbosity > 0:
        log_level = logging.DEBUG
    elif verbosity < 0:
        log_level = logging.ERROR
    else:
        log_level = logging.INFO
    if log_level == logging.DEBUG:
        log_format = '[ %(asctime)s | %(name)s | %(levelname)s ] %(message)s'
    else:
        log_format = '%(message)s'
    logging.basicConfig(
        level=log_level, format=log_format, datefmt='%d-%m-%Y %H:%M:%S')


def get_tool_command_line(args: Args) -> List[str]:
    tool = Tool(args.command)
    if command_line_str := args.custom_tool:
        command_line = process_command_line(command_line_str)
    else:
        provider: Optional[Provider] = None
        if args.use_pip_tools:
            provider = Provider.PIP_TOOLS
        elif args.use_uv:
            provider = Provider.UV
        if provider:
            command_line_str = provider.tools[tool]
            command_line, version = check_tool_version(command_line_str)
        else:
            command_line, version = find_tool(tool)
        log.debug('using %s %s', command_line, version)
    return command_line
